/*
 * Copyright (c) 2010, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using System.Collections;
using System.Text.RegularExpressions;
using System.IO;
using System.Threading;
using System.Runtime.InteropServices;

namespace MobiServer.Devices.DMX
{
    public partial class ControlPanel : MobiServer.ControlBase
    {

        #region DLL & Constans for DasHard
        [DllImport("DasHard2006VB", SetLastError = true)]

        // static extern unsafe int DasHardCommand(int command, int param, byte* dataArray);
        static extern unsafe int DasUsbCommand(int command, int param, byte* dataArray);

        const int DHC_INIT      =    9;     // COMMAND
        const int DHC_EXIT      =   10;     // COMMAND

        const int DHC_OPEN      =   1; 	    // COMMAND
        const int DHC_CLOSE     =   2; 	    // COMMAND
        const int DHC_DMXOUTOFF =	3; 	    // COMMAND
        const int DHC_DMXOUT    =	4;      // COMMAND
        const int DHC_PORTREAD  =	5; 	    // COMMAND
        const int DHC_PORTCONFIG= 	6; 	    // COMMAND
        const int DHC_VERSION 	 =  7; 	    // COMMAND
        const int DHC_DMXIN 	=   8;  	// COMMAND
        const int DHC_RESET 	=   11; 	// COMMAND
        const int DHC_DEBUG_OPEN  =	12; 	// COMMAND
        const int DHC_DEBUG_CLOSE = 13; 	// COMMAND
        const int DHC_WRITEMEMORY =	21; 	// COMMAND
        const int DHC_READMEMORY  =	22; 	// COMMAND
        const int DHC_SIZEMEMORY  = 23;  	// COMMAND
        const int DHC_TRANSPORT =	30;  	// COMMAND

        const int DHP_TRANSPORT_MODEALW =	1;	// PARAM
        const int DHP_TRANSPORT_MODEALW32 =	2; 	// PARAM
        const int DHP_TRANSPORT_MODEOPT =	3; 	// PARAM (DEFAULT)
        const int DHP_TRANSPORT_MODEOPT32 =	4; 	// PARAM

        const int DHC_SERIALNREAD  =	47;  	// COMMAND
        const int DHC_SERIALNWRITE =	48; 	// COMMAND

        const int DHC_SIUDI0 = 0;
        const int DHC_SIUDI1 = 100;
        const int DHC_SIUDI2 = 200;
        const int DHC_SIUDI3 = 300;
        const int DHC_SIUDI4 = 400;
        const int DHC_SIUDI5 = 500;
        const int DHC_SIUDI6 = 600;
        const int DHC_SIUDI7 = 700;
        const int DHC_SIUDI8 = 800;
        const int DHC_SIUDI9 = 900;	

        const int DHE_OK            =	1;   	    // RETURN NO ERROR
        const int DHE_NOTHINGTODO 	=   2;  	    // RETURN NO ERROR
        const int DHE_ERROR_COMMAND =  -1;   	    // RETURN ERROR
        const int DHE_ERROR_NOTOPEN =  -2;   	    // RETURN ERROR
        const int DHE_DMXOUT_PACKWRITE = -1000; 	// RETURN ERROR
	    const int ERROR_ACCESS_DENIED  = -1005; 	// ERROR_ACCESS_DENIED
	    const int ERROR_CRC            = -1023; 	// ERROR_CRC
        const int DHE_DMXOUT_PACKREAD = -1100; 	    // RETURN ERROR

        #endregion

        #region Constants/Variables
        Hashtable controlHolder = new Hashtable();   
        DMX.ControlDevice targetControl = null;
        int maxScreenHeight = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height - 50;
        int maxScreenWidth = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Width - 50;
        int controlWidthSum = 0;

        public MobiServer.ServerEventHandler serverWriteHandler;
        MobiServer.TCPServerControl serverControl = null;

        const string STR_OPEN_OK = "DMXfoCXւ̐ڑɐ܂";
        const string STR_OPEN_NG = "DMXfoCXւ̐ڑɎs܂";
        const string STR_CLOSE = "DMXfoCXւ̐ڑI܂";

        const string ERROR_NONAME = "DMXfoCXw肵ĂD";
        const string ERROR_EXIST_NAME = "DMXfoCXɑ݂Ă܂Dʂ̖Ow肵ĂD";
        const string ERROR_EXIST_CHANNEL = "DMX`ldĂ܂DݒmFĂD";
        const string ERROR_INVALID_NAME = "DMXfoCXɁu\\ / : * ! \" |v͎gpł܂Dʂ̖Ow肵ĂD";
        const string ERROR_RESERVED_NAME = "DMXfoCXɁuDirectv͎gpł܂Dʂ̖Ow肵ĂD";

        const string ERROR_INIT = "DMX DLL̏Ɏs܂D" + 
                                    "hCõCXg[󋵂mFĂD"+
                                    "vOI܂D";

        const int DMX_CHANNEL_MAX = 128;
        byte[] dmxDataAll = new byte[DMX_CHANNEL_MAX];
        byte[] dmxDataAllPre = new byte[DMX_CHANNEL_MAX];
        bool dmxOpenFlag = false;



        #endregion

        #region Constructor/Destructor

        public ControlPanel()
        {
            InitializeComponent();
           
            if (!InitDMX())
            {
                MessageBox.Show(ERROR_INIT, "Error");
                return;
            }
        }

        public ControlPanel(MobiServer.TCPServerControl server, MobiServer.MessageControl message)
            : base(message)
        {
            InitializeComponent();
            if (!InitDMX())
            {
                MessageBox.Show(ERROR_INIT, "Error");
                return;
            }
            
            serverWriteHandler += new MobiServer.ServerEventHandler(server.OnServerWriteEvent);
            server.AddReadHandler(new MobiServer.ServerEventHandler(OnServerReadEvent));
            this.serverControl = server;
        }

        private void Panel_Load(object sender, EventArgs e)
        {
            //panelSizePre = flowLayoutPanelDevice.Size;
            //controlWidthSum = flowLayoutPanelDevice.Width;
            LoadSetting();

            if (checkBoxAutoStart.Checked)
            {
                OpenDMX();
            }

            try
            {
                Hashtable h = ControlDevice.LoadDeviceAll(MessageControl);
                foreach (ControlDevice d in h.Values)
                {
                    StartControl(d);
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }

        void ChangeGUIEnabled(bool flag)
        {
        }

        unsafe bool InitDMX()
        {
            int result = DasUsbCommand(DHC_INIT, 0, (byte*)IntPtr.Zero.ToPointer());

            if (result > 0)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        unsafe void Finalize()
        {
            int result = 0;
            CloseDMX();
            timer1.Stop();
            result = DasUsbCommand(DHC_EXIT, 0, (byte*)IntPtr.Zero.ToPointer());
            return;
        }

        #endregion

        #region Control DMX Hardware

        private void buttonOpen_Click(object sender, EventArgs e)
        {
            OpenDMX();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            CloseDMX();
        }

        void OpenDMX()
        {
            buttonOpen.Enabled = false;
            try
            {
                if (SendDMXOpen())
                {
                    buttonClose.Enabled = true;
                    PrintMessage(STR_OPEN_OK);
                    timer1.Start();
                }
                else
                {
                    PrintError(STR_OPEN_NG);
                    buttonOpen.Enabled = true;
                    buttonClose.Enabled = false;
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
                buttonOpen.Enabled = true;
                buttonClose.Enabled = false;
            }
        }

        void CloseDMX()
        {
            buttonClose.Enabled = false;
            try
            {
                SendDMXClose();
                buttonOpen.Enabled = true;
                PrintMessage(STR_CLOSE);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            finally
            {
                buttonOpen.Enabled = true;
                timer1.Stop();
            }
        }

        public unsafe bool SendDMXOpen()
        {
            int result = 0;
            result = DasUsbCommand(DHC_OPEN, 0, (byte *)IntPtr.Zero.ToPointer());
            if (result >= 0)
            {
                result = DasUsbCommand(DHC_DMXOUTOFF, 0, (byte*)IntPtr.Zero.ToPointer());
                dmxOpenFlag = true;
                return true;
            }
            else
            {
                dmxOpenFlag = false;
                return false;
            }
        }

        public unsafe void SendDMXClose()
        {
            int result;
            result = DasUsbCommand(DHC_CLOSE, 0, (byte*)IntPtr.Zero.ToPointer());
            dmxOpenFlag = false;
            return;
        }

        
        public unsafe bool SendDMXData(int channel, byte data)
        {
            int result = 0;
            if (dmxOpenFlag && channel < dmxDataAll.Length)
            {
                dmxDataAll[channel] = data;

                fixed (byte* p = dmxDataAll)
                {
                    result = DasUsbCommand(DHC_DMXOUT, dmxDataAll.Length, p);
                }

                if (result >= 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return false;
            }
        }

        public unsafe bool SendDMXDataAuto()
        {
            int result = 0;
            if (dmxOpenFlag)
            {
                 fixed (byte* p = dmxDataAll)
                {
                    result = DasUsbCommand(DHC_DMXOUT, dmxDataAll.Length, p);
                }

                if (result >= 0)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return false;
            }
        }

        void OnDMXSendEvent(object sender, EventArgsChannel e)
        {
            try
            {
                SendDMXData(e.Channel, e.Value);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            SendDMXDataAuto();
        }

        #endregion

        #region Server In/Out
       
        public void OnServerReadEvent(object sender, MobiServer.ServerEventArgs e)
        {
            try
            {
                DataDevice d = DataDevice.Parse(e.Args);
                
                if (d != null)
                {
                    if (d.State == DataDevice.States.Attach)
                    {
                        StartControl(new ControlDevice(MessageControl, d.Type, d.ChannelBase, d.ChannelMax));
                    }
                }

            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }
        
        #endregion

        #region Add Controls

        public void StartControl(ControlDevice control)
        {
            serverControl.AddReadHandler(new MobiServer.ServerEventHandler(control.OnServerReadEvent));
            control.serverWriteHandler += new MobiServer.ServerEventHandler(serverControl.OnServerWriteEvent);
            control.dmxWriteHandler += new EventHandlerChannel(OnDMXSendEvent);
            AddControl(control);
            control.Dettach(false); 
        }

        internal void WriteServerSockets(string[] args)
        {
            serverControl.WriteAllSockets(args);
        }
        
        public void AddControl(ControlDevice control)
        {
            try
            {
                bool controlExisted = false;
                foreach (ControlDevice c in controlHolder.Values)
                {
                    if (control.CheckDeviceNameExist(c.DeviceName))
                    {
                        int num = this.Controls.IndexOf(c);
                        this.Controls[num].Enabled = true;
                        this.Refresh();
                        controlExisted = true;
                        break;
                    }
                }

                if (!controlExisted)
                {
                    targetControl = control;
                    controlHolder.Add(controlHolder.Count, control);
                    this.Invoke(new EventHandler(AddControl));
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
           return;


        }

        void AddControl(object sender, EventArgs e)
        {
            flowLayoutPanelDevice.SuspendLayout();
            flowLayoutPanelDevice.Controls.Add(targetControl);
            //Auto-layout based on screen width
            if (controlWidthSum + targetControl.Width > maxScreenWidth)
            {
                ResizePanelHeight();

                if (targetControl.Width > flowLayoutPanelDevice.Width)
                {
                    ResizePanelWidth(targetControl.Width - flowLayoutPanelDevice.Width);
                }

                controlWidthSum = targetControl.Width;
            }
            else
            {
                if (controlWidthSum + targetControl.Width > flowLayoutPanelDevice.Width)
                {
                    ResizePanelWidth(targetControl.Width);
                }
                controlWidthSum += targetControl.Width;
            }

            /*
           //Auto-layout based on window width
           if (controlWidthSum + targetControl.Width > flowLayoutPanelDevice.Width)
           {
               ResizePanelHeight();

               if (targetControl.Width > flowLayoutPanelDevice.Width)
               {
                   ResizePanelWidth();
               }

               controlWidthSum = targetControl.Width;
           }
           else
           {
               controlWidthSum += targetControl.Width;
           }*/

            flowLayoutPanelDevice.ResumeLayout(true);
        }


        void ResizePanelHeight()
        {
            int targetHeight = targetControl.Height + 10;

            if ((this.Height + this.Location.Y + targetHeight) <= maxScreenHeight)
            {
                flowLayoutPanelDevice.Height += targetHeight;
                this.Height += targetHeight;
            }
            else
            {
                flowLayoutPanelDevice.Height += targetHeight;
                this.Height = maxScreenHeight - this.Location.Y;
            }

        }

        void ResizePanelWidth(int targetWidth)
        {
            //int targetWidth = targetControl.Width - flowLayoutPanelDevice.Width;
            //int targetWidth = targetControl.Width;

            if ((this.Width + targetWidth + this.Location.X) <= maxScreenWidth)
            {
                flowLayoutPanelDevice.Width += targetWidth;
                this.Width += targetWidth;
            }
            else
            {
                flowLayoutPanelDevice.Width += targetWidth;
                this.Width = maxScreenWidth - this.Location.X;
            }
        }


        private void buttonAttach_Click(object sender, EventArgs e)
        {
            try
            {
                string deviceName = textBoxDeviceName.Text;
                int channelBase = (int) numericUpDownChannelBase.Value;
                int channelMax = (int) numericUpDownChannelCount.Value;

                if (deviceName == "")
                {
                    PrintError(ERROR_NONAME);

                }
                else if (CheckDeviceNameExist(deviceName))
                {
                    PrintError(ERROR_EXIST_NAME);
                }
                else if (CheckDeviceChannelExist(channelBase, channelMax))
                {
                    PrintError(ERROR_EXIST_CHANNEL);
                }
                else if (CheckInvalidName(deviceName))
                {
                    PrintError(ERROR_INVALID_NAME);
                }
                else if (Regex.Match(deviceName, "Direct", RegexOptions.IgnoreCase).Success)
                {
                    PrintError(ERROR_RESERVED_NAME);
                }
                else
                {
                    StartControl(new ControlDevice(MessageControl, deviceName, channelBase, channelMax));
                    DataDevice d = new DataDevice(DataDevice.States.Attach, deviceName, channelBase, channelMax);
                    string[] args = d.ToArray();
                    serverWriteHandler.Invoke(this, new MobiServer.ServerEventArgs(args));
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.ToString());
            }
        }

        
        bool CheckDeviceNameExist(string deviceName)
        {
            bool controlExisted = false;
            foreach (ControlDevice c in controlHolder.Values)
            {

                if (c.CheckDeviceNameExist(deviceName))
                {
                    controlExisted = true;
                    break;
                }
            }
            return controlExisted;
        }

        bool CheckDeviceChannelExist(int channelBase, int channelMax)
        {
            bool controlExisted = false;
            foreach (ControlDevice c in controlHolder.Values)
            {

                if (c.CheckDeviceChannelExist(channelBase, channelMax))
                {
                    controlExisted = true;
                    break;
                }
            }
            return controlExisted;
        }

       
        bool CheckInvalidName(string name){
            bool invalid = false;

            foreach (char c in Path.GetInvalidFileNameChars())
            {
                if (name.Contains(c.ToString()))
                {
                    invalid = true;
                    break;
                }
            }
            return invalid;
        }
        #endregion

        #region Save & Load

        public void SaveControlAll()
        {
            foreach (ControlDevice control in controlHolder.Values)
            {
                control.SaveSettingDevice();
            }
        }

        public string GetFilename()
        {
            string file = System.Windows.Forms.Application.StartupPath + @"\" + SETTING_DIRECTORY + @"\" + this.Name + ".txt";
            return file;
        }

        public void SaveSetting()
        {
            StreamWriter fileStream = null;
            string file = GetFilename();
           
            try
            {
                fileStream = new StreamWriter(file, false);
                fileStream.WriteLine("{0}\t{1}", checkBoxAutoStart.Checked, numericUpDownChannelMax.Value);
                fileStream.WriteLine("{0}\t{1}\t{2}", textBoxDeviceName.Text, numericUpDownChannelBase.Value, numericUpDownChannelCount.Value);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        public void LoadSetting()
        {
            StreamReader fileStream = null;
            string file = GetFilename();
            string line = "";
            string[] args = null;

            try
            {
                if (!File.Exists(file))
                    return;

                fileStream = new StreamReader(file);

                line = fileStream.ReadLine();
                args = line.Split("\t".ToCharArray());
                checkBoxAutoStart.Checked = bool.Parse(args[0]);
                numericUpDownChannelMax.Value = int.Parse(args[1]);

                line = fileStream.ReadLine();
                args = line.Split("\t".ToCharArray());
                textBoxDeviceName.Text = args[0];
                numericUpDownChannelBase.Value = int.Parse(args[1]);
                numericUpDownChannelCount.Value = int.Parse(args[2]);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        #endregion

      
      
      
    }
}

