using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

using System.Text.RegularExpressions;
using System.IO;
using System.Collections;

namespace MobiServer.Devices.DMX.Control
{

    public partial class Base : MobiServer.ControlBase
    {
        #region Constants/Variables
        protected char houseCode;
        protected int unitCode;
        protected string deviceType;

        public Data.ReadHandler serialReadHandler;
        public Data.WriteHandler serialWriteHandler;

        protected const string SETTING_DIR = "setting";

        bool commandSendFlag = true;
        #endregion

        #region Constructor/Destructor
        public Base()
        {
            InitializeComponent();
        }

        public Base(char houseCode, int unitCode, string deviceType)
            : base(Data.Base.Header, GetLabel(houseCode, unitCode, deviceType))
        {
            InitializeComponent();
            this.houseCode = houseCode;
            this.unitCode = unitCode;
            this.deviceType = deviceType;

            LoadSetting();

            switch (deviceType)
            {
                case Data.Types.Appliance:
                    groupBoxDim.Visible = false;
                    groupBoxPower.Location = new Point(groupBoxPower.Location.X + (int)(groupBoxDim.Width / 2),
                                                        groupBoxPower.Location.Y);
                    break;
            }
        }
        #endregion

        #region Properties
        public char HouseCode
        {
            get { return houseCode; }
        }

        public int UnitCode
        {
            get { return unitCode; }
        }

        public string DeviceType
        {
            get { return deviceType; }
        }
        #endregion

        #region Attach/Dettach

        private void buttonDettach_Click(object sender, EventArgs e)
        {
            Dettach(this.groupBox1.Enabled);
            if (groupBox1.Enabled)
            {
                CreateServerWriteEvent(Data.States.Attach);
            }
            else
            {
                CreateServerWriteEvent(Data.States.Dettach);
            }
        }

        public void Dettach(bool flag)
        {
            if (flag)
            {
                buttonDettach.Text = "^";
                this.groupBox1.Enabled = false;
                this.SaveSetting();
            }
            else
            {
                this.groupBox1.Enabled = true;
                buttonDettach.Text = "x";
            }
        }

        #endregion

        #region X10 In/Out
        public void SendOn()
        {
            SendCommand(houseCode, unitCode, Data.Commands.ON, 1);
        }

        public void SendOff()
        {
            SendCommand(houseCode, unitCode, Data.Commands.OFF, 1);
        }

        public void SendDim(int cnt)
        {

            SendCommand(houseCode, unitCode, Data.Commands.DIM, cnt);
        }

        public void SendBright(int cnt)
        {
            SendCommand(houseCode, unitCode, Data.Commands.BRIGHT, cnt);
        }

        void SendCommand(char houseCode, int unitCode, byte functionCode, int repeatCount)
        {
            Data.X10EventArgs e = new Data.X10EventArgs(houseCode, unitCode, functionCode, repeatCount);
            if (serialWriteHandler != null)
            {
                ChangeGUIEnabled(false);
                serialWriteHandler(this, e);
            }
        }

        public void OnSerialReadEvent(object sender, Data.X10EventArgs e)
        {
            if (CheckAddress(e.HouseCode, e.UnitCode))
            {
                commandSendFlag = false;
                switch (e.FunctionCode)
                {
                    case Data.Commands.ON:
                        ChangeGUIEnabled(true);
                        trackBarValue.Value = trackBarValue.Maximum;
                        radioButtonOn.Checked = true;
                        break;
                    case Data.Commands.OFF:
                        ChangeGUIEnabled(true);
                        trackBarValue.Value = trackBarValue.Minimum;
                        radioButtonOff.Checked = true;
                        break;
                    case Data.Commands.DIM:
                        ChangeGUIEnabled(true);
                        if (radioButtonOn.Checked)
                        {
                            if (trackBarValue.Value > trackBarValue.Minimum)
                                trackBarValue.Value -= 1;
                        }
                        break;
                    case Data.Commands.BRIGHT:
                        ChangeGUIEnabled(true);
                        if (radioButtonOn.Checked)
                        {
                            if (trackBarValue.Value < trackBarValue.Maximum)
                                trackBarValue.Value += 1;
                        }
                        else
                        {
                            radioButtonOn.Checked = true;
                            trackBarValue.Value = trackBarValue.Maximum;
                        }
                        break;
                }
                commandSendFlag = true;
            }
            Application.DoEvents();
        }

        private void radioButtonOn_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonOn.Checked)
            {
                if (commandSendFlag)
                    SendOn();
            }
        }

        private void radioButtonOff_CheckedChanged(object sender, EventArgs e)
        {
            if (radioButtonOff.Checked)
            {
                if (commandSendFlag)
                    SendOff();
            }
        }

        private void buttonBright_Click(object sender, EventArgs e)
        {
            if (commandSendFlag)
                SendBright(5);
        }

        private void buttonDark_Click(object sender, EventArgs e)
        {
            if (commandSendFlag)
                SendDim(5);
        }
        #endregion

        #region Server In/Out

        override public void OnServerReadEvent(object sender, MobiServer.ServerEventArgs e)
        {
            try
            {
                
                if (CheckAddress(e.Args))
                {
                    Data.Base b = Data.Base.Parse(e.Args);

                    if (b != null)
                    {
                        if (b.State == Data.States.Dettach)
                        {
                            Dettach(true);
                        }
                        else if (b.State == Data.States.Out)
                        {

                            if (b.Type == Data.Types.Power)
                            {
                                if (b.Data == 1)
                                {
                                    radioButtonOn.Checked = true;
                                }
                                else
                                {
                                    radioButtonOff.Checked = true;
                                }
                            }
                            else if (b.Type == Data.Types.Bright)
                            {
                                if (b.Data > 0)
                                {
                                    SendBright(b.Data);
                                }
                                else
                                {
                                    SendDim(Math.Abs(b.Data));
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }

        void CreateServerWriteEvent(string state)
        {
            switch (state)
            {
                case Data.States.Attach:
                case Data.States.Dettach:
                    Data.Base b = new Data.Base(state, this.houseCode, this.unitCode, this.deviceType);
                    string[] args = b.ToArray();
                     this.CreateServerWriteEvent(args);
                    break;
            }
        }

        #endregion

        #region Save / Load
        string GetFilename()
        {
            string str = System.Windows.Forms.Application.StartupPath + @"\" + SETTING_DIR + @"\" +
                                  String.Format("X10_{0}_{1}_{2}.txt", houseCode, unitCode, deviceType);
            return str;
        }

        public static Hashtable LoadX10All()
        {
            Hashtable holder = new Hashtable();

            string[] files = Directory.GetFiles(System.Windows.Forms.Application.StartupPath + @"\" + SETTING_DIR);

            foreach (string file in files)
            {
                string f = Path.GetFileNameWithoutExtension(file);
                Match m = Regex.Match(f, @"^X10_(\w)_(\d+)_(\w+)");
                if (m.Success)
                {
                    char hcode = char.Parse(m.Groups[1].ToString());
                    int ucode = int.Parse(m.Groups[2].ToString());
                    string dtype = m.Groups[3].ToString();
                    Base control = new Base(hcode, ucode, dtype);

                    //if (control.Enabled)
                    if (control.GetEnabled())
                        holder.Add(holder.Count, control);
                }
            }

            return holder;
        }

        public void SaveSetting()
        {
            StreamWriter fileStream = null;
            string file = GetFilename();
            try
            {
                fileStream = new StreamWriter(file, false);
                fileStream.WriteLine("{0}", this.GetEnabled());
                fileStream.WriteLine("{0}\t{1}", radioButtonOn.Checked, radioButtonOff.Checked);
                fileStream.WriteLine("{0}", trackBarValue.Value);
            }
            catch (Exception ex)
            {
                PrintMessage(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        public void LoadSetting()
        {
            StreamReader fileStream = null;
            string file = GetFilename();
            string line = "";
            string[] args = null;

            try
            {
                if (!File.Exists(file))
                    return;

                fileStream = new StreamReader(file);

                line = fileStream.ReadLine();
                this.groupBox1.Enabled = bool.Parse(line);
                line = fileStream.ReadLine();
                args = line.Split("\t".ToCharArray());
                radioButtonOn.Checked = bool.Parse(args[0]);
                radioButtonOff.Checked = bool.Parse(args[1]);
                line = fileStream.ReadLine();
                trackBarValue.Value = int.Parse(line);

            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        #endregion

        #region Utility / Others

        void ChangeGUIEnabled(bool flag)
        {
            groupBoxPower.Enabled = flag;
            if (deviceType == Data.Types.Lamp)
                groupBoxDim.Enabled = flag;
            else
                groupBoxDim.Enabled = false;
        }

        static string GetLabel(char houseCode, int unitCode, string deviceType)
        {
            string str = String.Format("{2}:{0}{1}", houseCode, unitCode, deviceType);
            return str;
        }

        public bool CheckAddress(char houseCode, int unitCode)
        {
            if (Regex.Match(houseCode.ToString(), this.houseCode.ToString(), RegexOptions.IgnoreCase).Success &&
                (unitCode == this.unitCode || unitCode == 0))
                return true;
            else
                return false;
        }

        bool CheckAddress(string[] args)
        {
            char houseCode = args[2].ToCharArray()[0];
            int unitCode = int.Parse(args[3]);

            bool flag = CheckAddress(houseCode, unitCode);

            return flag;
        }

        public void SetDeviceType(string deviceType)
        {
            if (deviceType != this.DeviceType)
            {
                this.deviceType = deviceType;
                groupBox1.Text = GetLabel(houseCode, unitCode, deviceType);
                switch (deviceType)
                {
                    case Data.Types.Appliance:
                        //groupBoxDim.Enabled = false;
                        groupBoxDim.Visible = false;
                        groupBoxPower.Location = new Point(groupBoxPower.Location.X + (int)(groupBoxDim.Width / 2),
                                                            groupBoxPower.Location.Y);
                        break;
                    case Data.Types.Lamp:
                        groupBoxDim.Visible = true;
                        groupBoxPower.Location = new Point(groupBoxPower.Location.X - (int)(groupBoxDim.Width / 2),
                                                            groupBoxPower.Location.Y);
                        break;
                }
            }
        }

        public bool GetEnabled()
        {
            return this.groupBox1.Enabled;
        }

        public virtual bool Start()
        {
            return false;
        }

        public void Stop()
        {
        }

        private void trackBarBright_Scroll(object sender, EventArgs e)
        {

        }

        #endregion


    }
}
